"""
@author Matt Murphy
"""
from hoplite.utils import server_logging
from .job import Job
from hoplite.exceptions import JobDoesNotExistError, JobPluginDoesNotExistError
import uuid

logger = server_logging.get_server_logger(__name__)


class JobManager(object):
    """
    Class used by the server to manage jobs
    """

    def __init__(self, plugin_manager):
        """
        Initialize with unique id for this instance
        and the configured plugin paths
        """
        self.plugin_manager = plugin_manager
        self.jobs = {}

    def available_job_plugins(self):
        """
        Returns a list of jobs that can be created
        """
        return self.plugin_manager.get_plugin_names()

    def all_jobs(self):
        """
        Get all jobs that have been created
        """
        return list(self.jobs.values())

    def get_job(self, job_uuid):
        job = self.jobs.get(job_uuid, None)
        if job is None:
            raise JobDoesNotExistError(
                "Job with UUID: {0} does not exist".format(job_uuid))
        return job

    def create_job(self, name, config, running=False, port=5000):
        """
        Stores information about job in the job dictionary.
        If running is true then starts the job.
        """
        module = self._get_plugin_with_name(name)
        job_uuid = str(uuid.uuid4())
        job_api_key = str(uuid.uuid4())
        # TODO: Try/Catch if run does not exist
        job = Job(
            job_uuid,
            name,
            config,
            job_api_key,
            entry_point_group_name=self.plugin_manager.entry_point_group_name,
            port=port)
        logger.debug("Creating Job:{0} UUID:{1}".format(name, job_uuid))
        self.jobs[job.uuid] = job
        if running:
            job.start()
        return job

    def _get_plugin_with_name(self, name):
        plugin = self.plugin_manager.get_plugin_module_by_name(name)
        if plugin is None:
            raise JobPluginDoesNotExistError(name)
        return plugin

    def _clear(self):
        logger.warning("Clearing all jobs")
        self.jobs = {}



"""
Idea taken from: https://github.com/mattupstate/overholt/blob/master/tests/utils.py
"""

from .status_updater import StatusUpdater, MockStatusUpdater
from .remote_job_manager import RemoteJobManager
from .remote_job import RemoteJob




DOWNLOAD_NETWORK_FOLDER_JOB_NAME = "hoplite.plugins.download_network_folder_job"
DOWNLOAD_FOLDER_FROM_FTP_JOB_NAME = "hoplite.plugins.download_folder_from_ftp_job"
UPLOAD_TO_NETWORK_FOLDER_JOB_NAME = "hoplite.plugins.upload_to_network_folder_job"
INSTALL_PYTHON_PACKAGE_JOB_NAME = "hoplite.plugins.install_python_package_job"
SYNC_PERFORCE_PATH_JOB_NAME = "hoplite.plugins.sync_perforce_path_job"


class DownloadFolderFromFtpJobConstants(object):
    DEST_ROOT = "dest_root"
    FTP_ROOT = "ftp_root"
    SERVER_ADDRESS = "server_address"
    SERVER_PORT = "server_port"
    USERNAME = "username"
    PASSWORD = "password"


class InstallPythonPackageJobConstants(object):
    LOCAL_PATH = "LOCAL_PATH"
    PIP_CMD = "PIP_CMD"


class SyncPerforcePathJobConstants(object):
    USER = "USER"
    CLIENT_WORKSPACE = "CLIENT_WORKSPACE"
    P4_HOST = "P4_HOST"
    P4_PORT = "P4_PORT"
    PERFORCE_PATH = "PERFORCE_PATH"
    FORCE = "FORCE"



import json

from httmock import urlmatch, HTTMock, response
from hoplite.client.status_updater import StatusUpdater
from hoplite.exceptions import JobDoesNotExistError
import unittest2


@urlmatch(path='\/jobs\/someuuid$')
def update_status(url, request):
    if json.loads(request.body)["api_key"] == "apikeyhere":
        return response(200)
    return response(404)

class TestStatusUpdater(unittest2.TestCase):
    def test_update(self):
        with HTTMock(update_status):
            status = StatusUpdater('localhost:5001', "someuuid", "apikeyhere")
            status.update({"some": "status"})

    def test_update_raises(self):
        with HTTMock(update_status):
            status = StatusUpdater('localhost:5001', "someuuid", "wrongapikey")
            self.assertRaises(JobDoesNotExistError, status.update, {"some": "status"})




"""
Role
====

:py:class:`hoplite.client.StatusUpdater` is one of the paramters of a job's
run() function.


:py:class:`hoplite.client.MockStatusUpdater` is for use during job development

Example
=======
While developing a job, you might want to invoke the job script from the
command line.  By using :py:class:`hoplite.client.MockStatusUpdater` you can
call the run method of your job and have all your status updates printed to the
console.

    ..  code-block:: python

        def run(config, status):
            # Some python code here
            pass

        if __name__ == "__main__":
            config = {}
            run(config, MockStatusUpdater())

API
===
"""
import requests
from hoplite.serializer import hoplite_dumps
from hoplite.exceptions import JobDoesNotExistError


class StatusUpdater(object):
    """
    Used to update the status of the job running on the daemon.
    This is passed in to the job's run() function when it is invoked.
    """
    def __init__(self, addr, uuid, api_key):
        self.addr = addr
        self._uuid = uuid
        self._daemon_addr = 'http://{0}'.format(self.addr)
        self._api_key = api_key
        self.status = {}

    def update(self, status):
        """
        Updates the job's status on the server that created it.

        :param dict status: The new status of the job
        """
        self.status = status
        body = hoplite_dumps({
            "api_key": self._api_key,
            "status": status
        })
        url = self._daemon_addr + '/jobs/{0}'.format(self._uuid)
        headers = {'content-type': 'application/json'}
        r = requests.put(url, data=body, headers=headers)
        if r.status_code == 404:
            raise JobDoesNotExistError


class MockStatusUpdater(object):
    """
    For use while developing a job
    """
    def __init__(self, addr="0.0.0.0:5000", uuid="{0}", api_key=""):
        self.addr = addr
        self._uuid = uuid
        self._daemon_addr = 'http://{0}'.format(self.addr)
        self._api_key = api_key
        self.status = {}

    def update(self, status):
        """
        Prints the status to stdout for debugging and sets the
        member variable status for use in testing status updates

        :param status: status dictionary
        """
        self.status = status
        print((self.status))



import logging
import pickle
import sys
import traceback

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class HopliteError(Exception):
    """
    Base exception for all Hoplite errors
    """
    pass


# TODO: Write unit tests for the logic in this class
class JobFailedError(HopliteError):
    """
    Raised when there is an "exception" key in a job's status on the client.

    This exception can contain nested information regarding other exceptions
    which preceded it. Specifically, in cases where jobs call other jobs which
    raise exceptions, stack trace and other important information for every
    intermediate job can be preserved in a single exception.  This is useful
    for debugging in complex situations which require chaining of multiple
    jobs.
    """
    def __init__(self, address, uuid, traceback_object, previous_exception):
        """
        :param String address: Address of the host that was running the job
        :param String uuid: uuid of the job that threw the exception
        :param [(String, String, String, String)] traceback_object: Traceback
            object for the exception which occurred
        :param previous_exception: The information from the previously-raised
            exceptions
        """
        self.addr = address
        self.uuid = uuid
        self.traceback_object = traceback_object
        self.previous_exception = previous_exception

    @property
    def message(self):
        """
        Ensure that full info about the exception will always be retrieved when
        viewing the message
        :return:
        """
        return self.__str__()

    def __str__(self):
        """
        Provides useful information when printing information about this
        exception. This will display the full traces for each job which was
        called leading up to the root exception.

        :return: A string containing a full representation of all information
            contained in this exception. This includes stack traces for all
            jobs called before the exception occurred.
        """
        def single_traceback(traceback_object):
            trace_str = ''
            for line in traceback.extract_tb(traceback_object):
                trace_str += '      File "{0}", line {1}, in {2}'.format(line[0], line[1], line[2]) + '\n'
                trace_str += '         {0}'.format(line[3]) + '\n'
            return trace_str

        output = ''
        output += 'Full traceback for all jobs descended from current job:\n'
        output += '   In job with UUID: {}'.format(self.uuid) + '\n'
        output += '   Running on machine: {}'.format(self.addr) + '\n'
        output += '      Traceback:\n'
        output += single_traceback(self.traceback_object)

        previous_exception = self.previous_exception
        while (previous_exception is not None) and (previous_exception.get('type', None) is None):
            output += 'In job with UUID: {}'.format(previous_exception.get('uuid')) + '\n'
            output += 'Running on machine: {}'.format(previous_exception.get('address')) + '\n'
            output += '   Traceback:\n'
            output += single_traceback(pickle.loads(previous_exception.get('traceback')))
            previous_exception = previous_exception.get('previous_exception', None)

        output += '   Root Error Type: {}'.format(previous_exception.get('type')) + '\n'
        output += '   Root Error Message: {}'.format(previous_exception.get('message'))
        return output

    def raise_remote_exception(self):
        """
        Handles JobFailedErrors by trying to raise the same exception which was
        raised on the remote machine. This is useful in allowing a client to
        operate on a remote machine more transparently (i.e. it seems more like
        the operations are occurring on the local machine, since the exceptions
        raised on the remote machine are also raised on the local machine).
        This functionality is probably vulnerable to many corner cases, and
        might not be particularly reliable.

        Examines the JobFailedError exception information to determine what
        exception was originally raised on the remote machine. If that
        exception exists in scope on this machine (i.e. a currently-loaded
        module defines that exception), then that exception will be raised on
        the local machine. Otherwise, a JobFailedError exception will be
        raised.

        This is called as part of the RemotableMetaClass and Remotify
        decorator, and should not be manually called in most circumstances

        :raises JobFailedError: When the original exception cannot be raised
            for any reason, then the JobFailedError from which this was called
            is reraised.
        :raises Various Other Errors: When a known exception occurred in the
            remote job, then that exception will be raised
        """

        # Get to bottom-most exception (the original exception that was thrown)
        previous_exception = self.previous_exception
        while (previous_exception is not None) and ('type' not in previous_exception):
            previous_exception = previous_exception.get('previous_exception', None)

        exception_object_string = previous_exception.get('exception_object', None)

        exception_to_raise = self

        try:
            if exception_object_string:
                exception_object = pickle.loads(exception_object_string)
                if exception_object is not None:
                    exception_to_raise = exception_object
        except Exception as e:  # If unable to raise original exception, then reraise self
            pass
        finally:
            logger.info('Exception raised from remote machine "{}". Local traceback:\n{}\n'
                        'Remote exception info:\n{}'.format(self.addr,
                                                            traceback.print_tb(sys.exc_info()[2]),
                                                            traceback.format_exception(
                                                                type(exception_to_raise),
                                                                exception_to_raise,
                                                                self.traceback_object
                                                            )))
            raise type(exception_to_raise)(exception_to_raise).with_traceback(self.traceback_object.tb_next.tb_next)


class JobPluginLoadError(HopliteError):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class JobNotStartedError(HopliteError):
    def __init__(self, uuid):
        self.msg = "Job UUID: {0} has not been started".format(uuid)

    def __str__(self):
        return self.msg


class JobAlreadyStartedError(HopliteError):
    def __init__(self, uuid):
        self.msg = "Job UUID: {0} you cannot start a job more than once".format(uuid)

    def __str__(self):
        return self.msg


class NotAuthorizedError(HopliteError):
    def __init__(self):
        pass

    def __str__(self):
        return "You are not authorized to perform this action"


class JobDoesNotExistError(HopliteError):
    def __init__(self, message="Job does not exist"):
        self.message = message

    def __str__(self):
        return self.message


class JobPluginDoesNotExistError(HopliteError):
    def __init__(self, name):
        self.job_plugin_name = name

    def __str__(self):
        return "Job plugin '{0}' does not exist".format(self.job_plugin_name)


class InternalServerError(HopliteError):
    def __init__(self, message="Something went wrong on the server"):
        self.message = message

    def __str__(self):
        return self.message


class TimeoutError(HopliteError):
    def __init__(self, uuid=0):
        self._uuid = uuid

    def __str__(self):
        return "Waiting for job {0} timed out".format(self._uuid)


class ConnectionError(HopliteError):
    def __init__(self, addr):
        self.addr = addr

    def __str__(self):
        return 'Hoplite could not be contacted on host "{0}". Check that IP ' \
               'address/hostname and port are correct, and (if using ' \
               'hostname) check that DNS server is correctly configured. ' \
               'Also check that Hoplite server is running on the ' \
               'host.'.format(self.addr)



import datetime
import logging
import logging.handlers
import os
import psutil
import random
import re
import string
import sys
import tempfile
import time
import unittest2

from hoplite.client.remote_job_manager import RemoteJobManager
from hoplite.exceptions import JobFailedError, TimeoutError
from hoplite.public_api import wait_for_hoplite

sys.path.append(os.path.realpath(__file__))
import remotable_test_resources.remotable_class as remotable_class
from remotable_test_resources.dummy_class import DummyClass
import remotable_test_resources.remotable_module as remotable_module


def start_hoplite_server(port_num):
    proc = psutil.Popen(
        '{} -c "import hoplite.main; hoplite.main.server_main([\'--port={}\'])"'.format(sys.executable, port_num)
    )
    wait_for_hoplite('localhost', port_num)
    return proc


def tear_down_hoplite(process):
    l = process.get_children()
    for child in l:
        child.terminate()
        child.wait()
    process.terminate()
    process.wait()
    time.sleep(.5)


class TestMetaClassInjection(unittest2.TestCase):
    def setUp(self):
        self.class_instance = remotable_class.TestClass()

    def tearDown(self):
        pass

    def test_meta_class_injects_functions(self):
        class_attribs = dir(self.class_instance)

        function_names = [
            'return_none',
            'return_number',
            'return_tuple',
            'return_single_list',
            'return_multiple_lists',
            'do_math',
            'pass_common_class',
            'pass_custom_class',
            'get_class_variables',
            'raise_type_error',
            'raise_custom_error',
            'create_file',
            'create_specified_file',
            'call_nested_function',
            'raise_private_error',
            'raise_public_error',
            'return_custom_exception',
            'long_job',
            'log_normal'
        ]

        for name in function_names:
            self.assertIn('remote_' + name, class_attribs)
            self.assertIn('remote_async_' + name, class_attribs)


class TestModuleInjection(unittest2.TestCase):
    def test_remotable_functions_injected_in_module(self):
        class_attribs = dir(remotable_module)

        function_names = [
            'return_none',
            'return_number',
            'return_tuple',
            'return_single_list',
            'return_multiple_lists',
            'do_math',
            'pass_common_class',
            'pass_custom_class',
            'raise_type_error',
            'raise_custom_error',
            'create_specified_file',
            'call_nested_function',
            'raise_private_error',
            'raise_public_error',
            'return_custom_exception',
            'long_job',
            'log_normal'
        ]

        for name in function_names:
            self.assertIn('remote_' + name, class_attribs)
            self.assertIn('remote_async_' + name, class_attribs)


class TestRemotableClassCapabilities(unittest2.TestCase):
    def setUp(self):
        """
        This function has to use subprocess instead of multiprocessing. Otherwise, tests fail
        with the error "daemonic processes are not allowed to have children". This way, hoplite is
        run in a separate process.
        :return:
        """
        self.proc = start_hoplite_server(5001)
        self.class_instance = remotable_class.TestClass('this_is@a#string', 12349)

    def tearDown(self):
        tear_down_hoplite(self.proc)
        time.sleep(.5)

    def test_return_none(self):
        ret = self.class_instance.remote_return_none('localhost:5001')
        self.assertIsNone(ret)

    def test_return_number(self):
        ret = self.class_instance.remote_return_number('localhost:5001')
        self.assertEqual(ret, 777)

    def test_return_tuple(self):
        ret = self.class_instance.remote_return_tuple('localhost:5001')
        self.assertTrue(type(ret) == tuple)
        self.assertTupleEqual(ret, ('All', 4, 1.11))

    def test_return_list(self):
        ret = self.class_instance.remote_return_single_list('localhost:5001')
        self.assertTrue(type(ret) == list)
        self.assertListEqual(ret, ['This', 'is', 'a', 'list'])

    def test_return_multiple_lists(self):
        ret = self.class_instance.remote_return_multiple_lists('localhost:5001')
        self.assertTrue(type(ret) == tuple)
        self.assertListEqual(ret[0], ['This', 'is', 1, 'list'])
        self.assertListEqual(ret[1], ['and', 'this', 'is', 'another', 'list'])

    def test_do_math(self):
        ret = self.class_instance.remote_do_math('localhost:5001', 17, 13.5)
        self.assertEqual(ret, 17 * 13.5)

    def test_pass_common_class(self):
        date = datetime.datetime.now()
        ret = self.class_instance.remote_pass_common_class('localhost:5001', date)
        self.assertEqual(ret[0], date.year)
        self.assertEqual(ret[1], date.month)
        self.assertEqual(ret[2], date.day)
        self.assertEqual(ret[3], date.hour)
        self.assertEqual(ret[4], date.minute)
        self.assertEqual(ret[5], date.second)
        self.assertEqual(ret[6], date.microsecond)

    def test_pass_custom_class(self):
        dummy = DummyClass(53, 35)
        ret = self.class_instance.remote_pass_custom_class('localhost:5001', dummy)
        self.assertEqual(ret, 88)

    def test_get_class_variables(self):
        ret = self.class_instance.remote_get_class_variables('localhost:5001')
        self.assertTrue(type(ret) == tuple)
        self.assertEqual(ret[0], 'this_is@a#string')
        self.assertEqual(ret[1], 12349 * 2)

    def test_raise_type_error(self):
        with self.assertRaises(TypeError) as error_context:
            self.class_instance.remote_raise_type_error('localhost:5001')
        e = error_context.exception
        self.assertEqual(e.message, '___Failure1')

    def test_raise_custom_error(self):
        with self.assertRaises(remotable_class.ExternalEmptyError) as error_context:
            self.class_instance.remote_raise_custom_error('localhost:5001')
        e = error_context.exception
        self.assertEqual(e.message, '___Failure2')

    def test_create_file(self):
        filename = self.class_instance.remote_create_file('localhost:5001')
        with open(filename) as fin:
            self.assertEqual(fin.read().strip(), self.class_instance.file_contents)
        os.remove(filename)

    def test_call_nested_function(self):
        with self.assertRaises(TypeError) as error_context:
            self.class_instance.remote_call_nested_function('localhost:5001')
        e = error_context.exception
        self.assertEqual(e.message, '___Failure1')

    def test_raise_private_error(self):
        """
        Test how the metaclass deals with exceptions that it cannot reraise. In this case,
        the MyPrivateError exception class is defined within the TestClass, and so it cannot
        be pickled. That means that it cannot be re-raised, and so a JobFailedError should
        be raised instead. However, we can still check that the MyPrivateError exception
        was raised on the server end.
        """
        with self.assertRaises(JobFailedError) as error_context:
            self.class_instance.remote_raise_private_error('localhost:5001')
        e = error_context.exception
        self.assertIn('Root Error Message: ___Failure3', e.__str__())
        match = re.search('Root Error Type: <class \'.*\.MyPrivateError\'>', e.__str__())
        self.assertIsNotNone(match)

    def test_raise_public_error(self):
        with self.assertRaises(remotable_class.ExternalCustomError) as error_context:
            self.class_instance.remote_raise_public_error('localhost:5001')
        self.assertEqual(error_context.exception.message, '___Failure4')

    def test_return_custom_exception(self):
        """
        Tests whether non-trivial class instances can be returned
        """
        ret = self.class_instance.remote_return_custom_exception('localhost:5001')
        self.assertIsInstance(ret, remotable_class.ExternalCustomError)
        self.assertEqual(ret.message, '___Failure4')

    def test_return_none_start_join(self):
        job = self.class_instance.remote_async_return_none('localhost:5001')
        job.start()
        ret = 'something_not_none'
        ret = job.join()
        self.assertIsNone(ret)

    def test_return_number_start_join(self):
        job = self.class_instance.remote_async_return_number('localhost:5001')
        job.start()
        ret = job.join()
        self.assertEqual(ret, 777)

    def test_create_specified_file_start(self):
        # This function is considered unsafe, but it should be fine in this case since there are no security concerns
        filename = tempfile.mktemp()
        job = self.class_instance.remote_async_create_specified_file('localhost:5001', filename, 'specified_contents')
        job.start()
        self.assertFalse(os.path.exists(filename))
        time.sleep(5)  # Give time to create file
        with open(filename) as fin:
            self.assertEqual(fin.read().strip(), 'specified_contents')
        os.remove(filename)
        job.join()

    def test_create_specified_file_default_contents_start(self):
        # This function is considered unsafe, but it should be fine in this case since there are no security concerns
        filename = tempfile.mktemp()
        job = self.class_instance.remote_async_create_specified_file('localhost:5001', filename)
        job.start()
        self.assertFalse(os.path.exists(filename))
        time.sleep(5)  # Give time to create file
        with open(filename) as fin:
            self.assertEqual(fin.read().strip(), 'default_contents')
        os.remove(filename)
        job.join()

    def test_derived_class(self):
        """
        Tests that the remotify decorator works properly when applied to classes which use inheritance
        """
        child_class_instance = remotable_class.ChildClass(30, 15)
        child_class_members = dir(child_class_instance)

        grandchild_class_instance = remotable_class.GrandchildClass(45, 30, 15)
        grandchild_class_members = dir(grandchild_class_instance)

        self.assertIn('remote_func_1', child_class_members)
        self.assertIn('remote_func_2', child_class_members)
        self.assertIn('remote_func_3', child_class_members)
        self.assertIn('remote_func_4', child_class_members)
        self.assertIn('remote_async_func_1', child_class_members)
        self.assertIn('remote_async_func_2', child_class_members)
        self.assertIn('remote_async_func_3', child_class_members)
        self.assertIn('remote_async_func_4', child_class_members)

        self.assertNotIn('remote_func_5', grandchild_class_members)
        self.assertNotIn('remote_func_6', grandchild_class_members)
        self.assertNotIn('remote_async_func_5', grandchild_class_members)
        self.assertNotIn('remote_async_func_6', grandchild_class_members)
        self.assertIn('func_1', grandchild_class_members)
        self.assertIn('func_2', grandchild_class_members)
        self.assertIn('func_3', grandchild_class_members)
        self.assertIn('func_4', grandchild_class_members)
        self.assertIn('remote_func_1', grandchild_class_members)
        self.assertIn('remote_func_2', grandchild_class_members)
        self.assertIn('remote_func_3', grandchild_class_members)
        self.assertIn('remote_func_4', grandchild_class_members)
        self.assertIn('remote_async_func_1', grandchild_class_members)
        self.assertIn('remote_async_func_2', grandchild_class_members)
        self.assertIn('remote_async_func_3', grandchild_class_members)
        self.assertIn('remote_async_func_4', grandchild_class_members)

        time.sleep(1)  # Extra delay, since this test seems sporadic
        self.assertEqual(child_class_instance.remote_func_3('localhost:5001', 3), 3 + 3 + 15 + 30)
        self.assertEqual(child_class_instance.remote_func_4('localhost:5001', 6), 4 + 6 + 15 + 30)
        self.assertEqual(grandchild_class_instance.remote_func_3('localhost:5001', 9), 3 + 9 + 15 + 30)
        self.assertEqual(grandchild_class_instance.remote_func_4('localhost:5001', 12), 4 + 12 + 15 + 30)


    def test_job_timeout(self):
        with self.assertRaises(TimeoutError):
            self.class_instance.remote_long_job('localhost:5001', remote_timeout=3)

    def test_job_does_not_timeout(self):
        self.class_instance.remote_long_job('localhost:5001', remote_timeout=12)

    def test_async_job_timeout(self):
        with self.assertRaises(TimeoutError):
            job = self.class_instance.remote_async_long_job('localhost:5001')
            job.start()
            job.join(3)

    def test_async_job_does_not_timeout(self):
        job = self.class_instance.remote_async_long_job('localhost:5001')
        job.start()
        job.join(12)

    ''' Still need to add support for static functions - may need to write another Hoplite plugin to handle them
    def test_call_static_function(self):
        ret = remotable_class.TestClass.remote_static_return_number('localhost:5001')
        self.assertEqual(ret, 999)
    '''


class TestRemotableModuleCapabilities(unittest2.TestCase):
    def setUp(self):
        self.proc = start_hoplite_server(5001)
        self.manager = RemoteJobManager("localhost", 5001)

    def tearDown(self):
        tear_down_hoplite(self.proc)
        time.sleep(.5)

    def test_return_none(self):
        ret = remotable_module.remote_return_none('localhost:5001')
        self.assertIsNone(ret)

    def test_return_number(self):
        ret = remotable_module.remote_return_number('localhost:5001')
        self.assertEqual(ret, 777)

    def test_return_tuple(self):
        ret = remotable_module.remote_return_tuple('localhost:5001')
        self.assertTrue(type(ret) == tuple)
        self.assertTupleEqual(ret, ('All', 4, 1.11))

    def test_return_list(self):
        ret = remotable_module.remote_return_single_list('localhost:5001')
        self.assertTrue(type(ret) == list)
        self.assertListEqual(ret, ['This', 'is', 'a', 'list'])

    def test_return_multiple_lists(self):
        ret = remotable_module.remote_return_multiple_lists('localhost:5001')
        self.assertTrue(type(ret) == tuple)
        self.assertListEqual(ret[0], ['This', 'is', 1, 'list'])
        self.assertListEqual(ret[1], ['and', 'this', 'is', 'another', 'list'])

    def test_do_math(self):
        ret = remotable_module.remote_do_math('localhost:5001', 17, 13.5)
        self.assertEqual(ret, 17 * 13.5)

    def test_pass_common_class(self):
        date = datetime.datetime.now()
        ret = remotable_module.remote_pass_common_class('localhost:5001', date)
        self.assertEqual(ret[0], date.year)
        self.assertEqual(ret[1], date.month)
        self.assertEqual(ret[2], date.day)
        self.assertEqual(ret[3], date.hour)
        self.assertEqual(ret[4], date.minute)
        self.assertEqual(ret[5], date.second)
        self.assertEqual(ret[6], date.microsecond)

    def test_pass_custom_class(self):
        dummy = DummyClass(53, 35)
        ret = remotable_module.remote_pass_custom_class('localhost:5001', dummy)
        self.assertEqual(ret, 88)

    def test_raise_type_error(self):
        with self.assertRaises(TypeError) as error_context:
            remotable_module.remote_raise_type_error('localhost:5001')
        e = error_context.exception
        self.assertEqual(e.message, '___Failure1')

    def test_raise_custom_error(self):
        with self.assertRaises(remotable_module.ExternalEmptyError) as error_context:
            remotable_module.remote_raise_custom_error('localhost:5001')
        e = error_context.exception
        self.assertEqual(e.message, '___Failure2')

    def test_call_nested_function(self):
        with self.assertRaises(TypeError) as error_context:
            remotable_module.remote_call_nested_function('localhost:5001')
        e = error_context.exception
        self.assertEqual(e.message, '___Failure1')

    def test_raise_private_error(self):
        """
        Test how the metaclass deals with exceptions that it cannot reraise. In this case,
        the MyPrivateError exception class is defined within the TestClass, and so it cannot
        be pickled. That means that it cannot be re-raised, and so a JobFailedError should
        be raised instead. However, we can still check that the MyPrivateError exception
        was raised on the server end.
        """
        with self.assertRaises(JobFailedError) as error_context:
            remotable_module.remote_raise_private_error('localhost:5001')
        e = error_context.exception
        self.assertIn('Root Error Message: ___Failure3', e.__str__())
        match = re.search('Root Error Type: <class \'.*\.MyPrivateError\'>', e.__str__())
        self.assertIsNotNone(match)

    def test_raise_public_error(self):
        with self.assertRaises(remotable_class.ExternalCustomError) as error_context:
            remotable_module.remote_raise_public_error('localhost:5001')
        self.assertEqual(error_context.exception.message, '___Failure4')

    def test_return_custom_exception(self):
        """
        Tests whether non-trivial class instances (such as exceptions) can be returned
        """
        ret = remotable_module.remote_return_custom_exception('localhost:5001')
        self.assertIsInstance(ret, remotable_class.ExternalCustomError)
        self.assertEqual(ret.message, '___Failure4')

    def test_return_none_start_join(self):
        job = remotable_module.remote_async_return_none('localhost:5001')
        job.start()
        ret = 'something_not_none'
        ret = job.join()
        self.assertIsNone(ret)

    def test_return_number_start_join(self):
        job = remotable_module.remote_async_return_number('localhost:5001')
        job.start()
        ret = job.join()
        self.assertEqual(ret, 777)

    def test_create_specified_file_start(self):
        # This function is considered unsafe, but it should be fine in this case since there are no security concerns
        filename = tempfile.mktemp()
        job = remotable_module.remote_async_create_specified_file('localhost:5001', filename, 'specified_contents')
        job.start()
        self.assertFalse(os.path.exists(filename))
        time.sleep(3)  # Give time to create file
        with open(filename) as fin:
            self.assertEqual(fin.read().strip(), 'specified_contents')
        os.remove(filename)
        job.join()

    def test_create_specified_file_default_contents_start(self):
        # This function is considered unsafe, but it should be fine in this case since there are no security concerns
        filename = tempfile.mktemp()
        job = remotable_module.remote_async_create_specified_file('localhost:5001', filename)
        job.start()
        self.assertFalse(os.path.exists(filename))
        time.sleep(3)  # Give time to create file
        with open(filename) as fin:
            self.assertEqual(fin.read().strip(), 'default_contents')
        os.remove(filename)
        job.join()

    def test_job_timeout(self):
        with self.assertRaises(TimeoutError):
            remotable_module.remote_long_job('localhost:5001', remote_timeout=3)

    def test_job_does_not_timeout(self):
        remotable_module.remote_long_job('localhost:5001', remote_timeout=12)

    def test_async_job_timeout(self):
        with self.assertRaises(TimeoutError):
            job = remotable_module.remote_async_long_job('localhost:5001')
            job.start()
            job.join(3)

    def test_async_job_does_not_timeout(self):
        job = remotable_module.remote_async_long_job('localhost:5001')
        job.start()
        job.join(12)


class TestRemotableClassClientLogging(unittest2.TestCase):
    def setUp(self):
        self.logger = logging.getLogger('hoplite.remote_enabler')
        self.logger_path = tempfile.mktemp()
        self.logger_handler = logging.FileHandler(self.logger_path)
        self.logger.addHandler(self.logger_handler)
        self.logger.setLevel(logging.DEBUG)

        self.proc = start_hoplite_server(5001)
        self.manager = RemoteJobManager("localhost", 5001)
        self.class_instance = remotable_class.TestClass('this_is@a#string', 12349)

    def tearDown(self):
        tear_down_hoplite(self.proc)
        self.logger_handler.close()
        self.logger.removeHandler(self.logger_handler)

    def test_logging__remote_func(self):
        ret = self.class_instance.remote_do_math('localhost:5001', 2, 3)
        self.assertEqual(ret, 6)

        with open(self.logger_path) as fin:
            all_lines = fin.readlines()
        self.assertEqual(all_lines[0].strip(), '"{0}" on target "{1}" with args: {2} and kwargs: {3}'.format(
            'do_math', 'localhost:5001', (2, 3), {}))
        self.assertEqual(all_lines[1].strip(), '"{0}" on target "{1}" returned {2}'.format('do_math', 'localhost:5001', 6))

    def test_logging__remote_async_func(self):
        job = self.class_instance.remote_async_do_math('localhost:5001', 2, 3)
        job.start()
        ret = job.join()

        self.assertEqual(ret, 6)

        with open(self.logger_path) as fin:
            all_lines = fin.readlines()

        self.assertEqual(all_lines[0].strip(), 'Creating job "{0}" on target "{1}" with args: {2} and kwargs: {3}'.format(
            'do_math', 'localhost:5001', (2, 3), {}))
        self.assertIsNotNone(r'Starting "do_math\(.*\)" on "localhost:5001"', all_lines[1].strip())
        self.assertEqual(all_lines[2].strip(), 'Joining "{0}" on "{1}"'.format('do_math', 'localhost:5001'))
        self.assertEqual(all_lines[3].strip(), '"{0}" on target "{1}" returned {2}'.format(
            'do_math', 'localhost:5001', 6))


class TestRemotableModuleClientLogging(unittest2.TestCase):
    def setUp(self):
        self.logger = logging.getLogger('hoplite.remote_enabler')
        self.logger_path = tempfile.mktemp()
        self.logger_handler = logging.FileHandler(self.logger_path)
        self.logger.addHandler(self.logger_handler)
        self.logger.setLevel(logging.DEBUG)

        self.proc = start_hoplite_server(5001)
        self.manager = RemoteJobManager("localhost", 5001)

    def tearDown(self):
        tear_down_hoplite(self.proc)
        self.logger_handler.close()
        self.logger.removeHandler(self.logger_handler)

    def test_logging__remote_module_func(self):
        ret = remotable_module.remote_do_math('localhost:5001', 2, 3)
        self.assertEqual(ret, 6)

        with open(self.logger_path) as fin:
            all_lines = fin.readlines()
        self.assertEqual(all_lines[0].strip(), '"{0}" on target "{1}" with args: {2} and kwargs: {3}'.format(
            'do_math', 'localhost:5001', (2, 3), {}))
        self.assertEqual(all_lines[1].strip(), '"{0}" on target "{1}" returned {2}'.format('do_math', 'localhost:5001', 6))

    def test_logging__remote_module_async_func(self):
        job = remotable_module.remote_async_do_math('localhost:5001', 2, 3)
        job.start()
        ret = job.join()

        self.assertEqual(ret, 6)

        with open(self.logger_path) as fin:
            all_lines = fin.readlines()
        self.assertEqual(all_lines[0].strip(), 'Creating job "{0}" on target "{1}" with args: {2} and kwargs: {3}'.format(
            'do_math', 'localhost:5001', (2, 3), {}))
        self.assertIsNotNone(r'Starting "do_math\(.*\)" on "localhost:5001"', all_lines[1].strip())
        self.assertEqual(all_lines[2].strip(), 'Joining "{0}" on "{1}"'.format('do_math', 'localhost:5001'))
        self.assertEqual(all_lines[3].strip(), '"{0}" on target "{1}" returned {2}'.format(
            'do_math', 'localhost:5001', 6))


class TestRemotableClassServerLogging(unittest2.TestCase):
    def setUp(self):
        self.proc = start_hoplite_server(5001)
        self.manager = RemoteJobManager("localhost", 5001)
        self.class_instance = remotable_class.TestClass('this_is@a#string', 12349)

    def tearDown(self):
        tear_down_hoplite(self.proc)

    def test_logging_log_normal(self):
        rand_string_1 = get_random_string(5, 10)
        rand_string_2 = get_random_string(5, 10)
        log_folder = r'C:\logs\hoplite\remoted_functions\tests\remotable_test_resources\remotable_class\TestClass'
        ret = self.class_instance.remote_log_normal('localhost:5001', rand_string_1, dummy_var_2=rand_string_2)
        latest_log_file = max(
            [os.path.join(log_folder, filename) for filename in os.listdir(log_folder)], key=os.path.getctime
        )
        self.assertEqual(ret, '{} + {}'.format(rand_string_1, rand_string_2))
        with open(os.path.join(log_folder, latest_log_file)) as fin:
            all_lines = fin.readlines()
        self.assertEqual(len(all_lines), 3)
        self.assertIn("Beginning execution of log_normal with args: ('{}',) and kwargs: {{'dummy_var_2': '{}'}}".format(
            rand_string_1, rand_string_2), all_lines[0]
        )
        self.assertIn('Logging in log_normal function', all_lines[1])
        self.assertIn(
            'Returning from log_normal with return value(s): {} + {}'.format(rand_string_1, rand_string_2), all_lines[2]
        )
        os.remove(latest_log_file)

    def test_logging_log_exception(self):
        log_folder = r'C:\logs\hoplite\remoted_functions\tests\remotable_test_resources\remotable_class\TestClass'
        with self.assertRaises(TypeError) as error_context:
            self.class_instance.remote_raise_type_error('localhost:5001')
        self.assertEqual(error_context.exception.message, '___Failure1')

        latest_log_file = max(
            [os.path.join(log_folder, filename) for filename in os.listdir(log_folder)], key=os.path.getctime
        )
        with open(os.path.join(log_folder, latest_log_file)) as fin:
            all_lines = fin.readlines()
        self.assertEqual(len(all_lines), 8)
        self.assertIn('Beginning execution of raise_type_error with args: () and kwargs: {}', all_lines[0])
        self.assertIn('An exception occurred', all_lines[1])
        self.assertIn('Traceback', all_lines[2])
        self.assertIn('in run', all_lines[3])
        self.assertIn('TypeError: ___Failure1', all_lines[7])
        os.remove(latest_log_file)

    def test_logging_nested_logs(self):
        log_folder = r'C:\logs\hoplite\remoted_functions\tests\remotable_test_resources\remotable_class\TestClass'
        self.class_instance.remote_log_nested_caller('localhost:5001')

        latest_log_file = max(
            [os.path.join(log_folder, filename) for filename in os.listdir(log_folder)], key=os.path.getctime
        )
        with open(os.path.join(log_folder, latest_log_file)) as fin:
            all_lines = fin.readlines()
        self.assertEqual(len(all_lines), 4)
        self.assertIn('Currently in class caller function', all_lines[1])
        self.assertIn('Currently in class callee function', all_lines[2])
        os.remove(latest_log_file)


class TestRemotableModuleServerLogging(unittest2.TestCase):
    def setUp(self):
        self.proc = start_hoplite_server(5001)
        self.manager = RemoteJobManager("localhost", 5001)

    def tearDown(self):
        tear_down_hoplite(self.proc)

    def test_logging_log_normal(self):
        rand_string_1 = get_random_string(5, 10)
        rand_string_2 = get_random_string(5, 10)
        log_folder = r'C:\logs\hoplite\remoted_functions\tests\remotable_test_resources\remotable_module'
        ret = remotable_module.remote_log_normal('localhost:5001', rand_string_1, dummy_var_2=rand_string_2)
        latest_log_file = max(
            [os.path.join(log_folder, filename) for filename in os.listdir(log_folder)], key=os.path.getctime
        )
        self.assertEqual(ret, '{} + {}'.format(rand_string_1, rand_string_2))
        with open(os.path.join(log_folder, latest_log_file)) as fin:
            all_lines = fin.readlines()
        self.assertEqual(len(all_lines), 3)
        self.assertIn("Beginning execution of log_normal with args: ('{}',) and kwargs: {{'dummy_var_2': '{}'}}".format(
            rand_string_1, rand_string_2), all_lines[0]
        )
        self.assertIn('Logging in log_normal function', all_lines[1])
        self.assertIn(
            'Returning from log_normal with return value(s): {} + {}'.format(rand_string_1, rand_string_2), all_lines[2]
        )
        os.remove(latest_log_file)

    def test_logging_log_exception(self):
        log_folder = r'C:\logs\hoplite\remoted_functions\tests\remotable_test_resources\remotable_module'
        with self.assertRaises(TypeError) as error_context:
            remotable_module.remote_raise_type_error('localhost:5001')
        self.assertEqual(error_context.exception.message, '___Failure1')

        latest_log_file = max(
            [os.path.join(log_folder, filename) for filename in os.listdir(log_folder)], key=os.path.getctime
        )
        with open(os.path.join(log_folder, latest_log_file)) as fin:
            all_lines = fin.readlines()
        self.assertEqual(len(all_lines), 8)
        self.assertIn('Beginning execution of raise_type_error with args: () and kwargs: {}', all_lines[0])
        self.assertIn('An exception occurred', all_lines[1])
        self.assertIn('Traceback', all_lines[2])
        self.assertIn('in run', all_lines[3])
        self.assertIn('TypeError: ___Failure1', all_lines[7])
        os.remove(latest_log_file)

    def test_logging_nested_logs(self):
        log_folder = r'C:\logs\hoplite\remoted_functions\tests\remotable_test_resources\remotable_module'
        remotable_module.remote_log_nested_caller('localhost:5001')

        latest_log_file = max(
            [os.path.join(log_folder, filename) for filename in os.listdir(log_folder)], key=os.path.getctime
        )
        with open(os.path.join(log_folder, latest_log_file)) as fin:
            all_lines = fin.readlines()
        self.assertEqual(len(all_lines), 4)
        self.assertIn('Currently in caller function', all_lines[1])
        self.assertIn('Currently in callee function', all_lines[2])
        os.remove(latest_log_file)


def get_random_string(min_length, max_length):
    return ''.join(random.choice(string.ascii_lowercase + string.ascii_uppercase + string.digits)
                   for _ in range(random.randint(min_length, max_length)))


import os
import platform
import shutil
import stat
import unittest2

from hoplite.builtin_plugins.constants import SyncPerforcePathJobConstants as KEYS
from hoplite.builtin_plugins.sync_perforce_path_job import run as sync_perforce_path
from hoplite.client.status_updater import MockStatusUpdater


class TestSyncPerforcePathJob(unittest2.TestCase):
    def setUp(self):
        self.opsys = platform.system().lower()
        if self.opsys == 'windows':
            self.workspace = "riormt_hoplite_test_windows"
            self.workspace_root = r"C:\riormt_hoplite_test_windows"
            self.export_location_path = "C:\\riormt_hoplite_test_windows\\microDAQ\\exportLocation"
        elif self.opsys == 'linux' or self.opsys == 'darwin':
            self.workspace = "riormt_hoplite_test_posix"
            self.workspace_root = "/tmp/riormt_hoplite_test_posix"
            self.export_location_path = "/tmp/riormt_hoplite_test_posix/microDAQ/exportLocation"
        self.status = MockStatusUpdater()

    def test_sync_works(self):
        try:
            synced_file = os.path.join(self.workspace_root, "microDAQ", "exportLocation")
            config = {
                KEYS.USER: "riormt",
                KEYS.P4_PORT: "penguin.natinst.com:1666",
                KEYS.PERFORCE_PATH: "//microDAQ/exportLocation",
                KEYS.CLIENT_WORKSPACE: self.workspace,
                KEYS.FORCE: True
            }
            sync_perforce_path(config, self.status)
            self.assertEqual(self.export_location_path, self.status.status["local_path"])
            self.assertTrue(os.path.exists(synced_file))
            self.assertTrue(self.status.status["succeeded"])
        finally:
            try:
                os.chmod(self.export_location_path, stat.S_IWRITE)
                shutil.rmtree(self.workspace_root)
            except OSError:
                self.fail("Directory was not synced from perforce...")

    def test_sync_errors(self):
        config = {
            KEYS.USER: "riormt",
            KEYS.P4_PORT: "penguin.natinst.com:1666",
            KEYS.PERFORCE_PATH: "//microDAQ",
            KEYS.CLIENT_WORKSPACE: self.workspace,
            KEYS.FORCE: True
        }
        sync_perforce_path(config, self.status)
        self.assertFalse(self.status.status["succeeded"])


from hoplite.utils import server_logging
import ftplib
import os
import socket
from ftplib import FTP
from hoplite.client.status_updater import MockStatusUpdater
from hoplite.builtin_plugins.constants import DownloadFolderFromFtpJobConstants as KEYS

logger = server_logging.get_job_logger(__name__)


def _get_files_in_dir(ftp_session, source, dest, status):
    filelist = ftp_session.nlst()
    for path in filelist:
        local_path = os.path.join(dest, path)
        remote_path = source + path
        try:
            # See if it's a directory
            new_remote_path = source + path + "/"
            ftp_session.cwd(source + path + "/")
            # It is! So we should create the directory on dest
            try:
                os.mkdir(local_path)
            except OSError:
                logger.debug("Dir: {0} already exists".format(local_path))
            logger.debug("Get Folder: {0}".format(new_remote_path))
            _get_files_in_dir(ftp_session, new_remote_path, local_path, status)
        except ftplib.error_perm:
            logger.debug("Get file: {0}".format(path))
            with open(local_path, "wb") as file_handle:
                ftp_session.retrbinary(
                    "RETR " + remote_path, file_handle.write)


def download_files(ftp_session, source, dest, status):
    # dest must already exist
    try:
        ftp_session.cwd(source)
    except ftplib.error_perm:
        # invalid entry (ensure input form: "/dir/folder/something/")
        msg = "Could not open the ftp directory: '{0}'".format(source)
        logger.error(msg)
        status.update({"error": msg})
        return
    if not os.path.isdir(dest):
        msg = "Local path does not exist: '{0}'".format(dest)
        logger.error(msg)
        status.update({"error": msg})
        return
    _get_files_in_dir(ftp_session, source, dest, status)


def run(config, status):
    """
    Job to recursively download a directory from an FTP server
    This will overwrite any files that are in the dest_root directory
    """
    ftp_addr = config.get(KEYS.SERVER_ADDRESS, "localhost")
    ftp_port = config.get(KEYS.SERVER_PORT, 21)
    ftp_root = config.get(KEYS.FTP_ROOT, "/")
    dest_root = config.get(KEYS.DEST_ROOT, "C:\\temp\\")
    user = config.get(KEYS.USERNAME, "")
    password = config.get(KEYS.PASSWORD, "")

    try:
        ftp_session = FTP()
        ftp_session.connect(ftp_addr, ftp_port)
        ftp_session.login(user, password)
    except socket.gaierror as e:
        status.update({"error": str(e)})
        logger.error(e)
        return
    logger.debug("connected {0}")
    download_files(ftp_session, ftp_root, dest_root, status)
    ftp_session.close()

if __name__ == "__main__":
    config_dict = {
        "server": "localhost",
        "ftp_root": "/",
        "dest_root": "C:\\temp\\",
    }
    run(config_dict, MockStatusUpdater())



from hoplite.builtin_plugins.utils.network_file_operations import download_network_folder
from hoplite.client.status_updater import MockStatusUpdater
import shutil
import os


def run(config, status):
    """
    Copies a folder from a network location and places it on the local disk
    This job is for Windows only (you better remember to escape your
    backslashes)
    **If overwrite is False local_path must not exist**
    Expected Config:
    {
        "network_path": "\\\\share\\something\\something\\something",
        "local_path": "C:\\Documents\\doesnotexist",
        "user": "DOMAIN\\username",
        "password": "password",
        "overwrite": False
    }
    """
    network_path = config.get("network_path", None)
    local_path = config.get("local_path", None)
    user = config.get("user", "")
    password = config.get("password", "")
    overwrite = config.get("overwrite", False)

    if network_path is None:
        status.update(
            {"success": False, "error": "network_path not defined in config"})
        return
    if local_path is None:
        status.update(
            {"success": False, "error": "local_path not defined in config"})
        return

    if os.path.isdir(local_path):
        if overwrite:
            try:
                shutil.rmtree(local_path)
            except OSError as e:
                status.update({"success": False, "error": e})
                return
        else:
            status.update(
                {
                    "success": False,
                    "error": "Directory already exists: {0}".format(
                        local_path)
                })
            return

    if network_path.find('\\\\', 0, 2) == 0:
        stdout = ""
        try:
            stdout = download_network_folder(
                network_path, local_path, user, password)
        except RuntimeError as e:
            status.update({"success": False, "error": e})
            return
        status.update({"success": True, "stdout": stdout})
        return
    status.update({"error": "Invalid network path"})



class DummyClass:
    def __init__(self, num_1, num_2):
        self.num_1 = num_1
        self.num_2 = num_2

    def add_numbers(self):
        return self.num_1 + self.num_2

# TODO: Create a new-style class (inherit from "object") and write tests for that as well


from argparse import ArgumentParser
import sys
import os
import shutil
import platform


# TODO: Modify this to support custom port numbers
def main(args=sys.argv):
    '''
    Calls a batch file to start hoplite-server on boot.
    '''
    parser = ArgumentParser(
        description='Installs the hoplite-server in the startup directory.')
    parser.add_argument('-d', '--disable', action='store_true',
                        help='Disables auto start on boot.',
                        dest='disable')
    if args == sys.argv:
        args = args[1:]
    args = parser.parse_args(args)
    args = vars(args)
    working_dir = os.path.dirname(os.path.realpath(__file__))
    batch = os.path.join(working_dir, 'hoplite.bat')
    dest = ''
    if platform.release() == 'XP':
        dest = os.path.join(os.environ['allusersprofile'], 'Start Menu',
                            'Programs', 'Startup', 'hoplite.bat')
    else:
        dest = os.path.join(os.environ['allusersprofile'], 'Microsoft',
                            'Windows', 'Start Menu', 'Programs', 'Startup',
                            'hoplite.bat')
    if args['disable']:
        os.remove(dest)
        print('Hoplite auto start successfully disabled.')
    else:
        shutil.copyfile(batch, dest)
        print('Hoplite successfully configured to start on boot.')



import unittest2
import pip
import types
import os
import sys
from tests.paths import TEST_ENTRY_POINT_PATH
from hoplite.api.root import reload_site_packages
from hoplite.plugin_manager import EntryPointManager
import pkg_resources


def run_pip(pip_args):
    try:
        pip.main(initial_args=pip_args)
    except TypeError:  # The virtualenv pip uses "args" for the keyword argument name, rather than "initial_args"
        pip.main(args=pip_args)


class TestEntryPointManager(unittest2.TestCase):
    @classmethod
    def setUpClass(cls):
        # Install hoplite so that its entry points are added
        pip_args = ["install", os.path.join(os.path.dirname(os.path.abspath(__file__)), '..')]
        run_pip(pip_args)
        reload_site_packages()

    @classmethod
    def tearDownClass(cls):
        pip_args = ["uninstall", "hoplite", "-y"]
        run_pip(pip_args)

    def setUp(self):
        self.manager = EntryPointManager()

    def test_get_names_with_default_group_name(self):
        names = self.manager.get_plugin_names()
        self.assertIn('hoplite.plugins.download_folder_from_ftp_job', names)
        self.assertIn('hoplite.plugins.download_network_folder_job', names)
        self.assertIn('hoplite.plugins.upload_to_network_folder_job', names)

    def test_get_names_of_hot_installed_package(self):
        pip_args = ["install", TEST_ENTRY_POINT_PATH]
        run_pip(pip_args)
        reload_site_packages()
        for syspath in sys.path:
            pkg_resources.working_set.add_entry(syspath)
        try:
            self.manager = EntryPointManager('entry.point.test')
            names = self.manager.get_plugin_names()
            self.assertIn("entry_point_1", names)
            self.assertIn("entry_point_2", names)
        except Exception, e:
            raise e
        finally:
            pip_args = ["uninstall", "entrypointtest", "-y"]
            run_pip(pip_args)

    def test_get_module_returns_module(self):
        for entry_point in pkg_resources.iter_entry_points(group='hoplite.jobs'):
            if entry_point.name == 'hoplite.plugins.download_folder_from_ftp_job':
                mod = entry_point.load()
        self.assertIsInstance(mod, types.ModuleType)


from setuptools import setup, find_packages
import os
import re


def hoplite_plugins():
    hoplite_top_level = os.path.split(__file__)[0]
    plugin_dir = os.path.join(hoplite_top_level, "hoplite", "builtin_plugins")
    plugins = []
    for directory, dirnames, filenames in os.walk(plugin_dir):
        for filename in filenames:
            match = re.search('(\w+_job)\.py$', filename)
            if match:
                fqdn = 'hoplite.plugins.{0}'.format(match.group(1))
                module_name = 'hoplite.builtin_plugins.{0}'.format(match.group(1))
                entry = '{0} = {1}'.format(fqdn, module_name)
                plugins.append(entry)
    return plugins

setup(
    name="hoplite",
    version="15.0.0.dev21",
    packages=find_packages(exclude='tests'),
    license='MIT',
    install_requires=['flask>=0.10.1',
                      'requests>=2.2.0',
                      'argparse>=1.1',
                      'pymongo>=3.0',
                      'tornado',
                      'tblib'],
    download_url = "https://gitlab.com/croningroup/chemputer/hoplite/-/archive/v0.1.0/hoplite-v0.1.0.tar.gz",
    entry_points={
        'console_scripts': [
            'hoplite-server = hoplite.main:server_main',
            'hoplite-client = hoplite.main:client_main',
            'hoplite-auto-start = hoplite.auto_start:main'
        ],
        'hoplite.jobs': hoplite_plugins()
    },
    include_package_data=True,
    package_data={
        '': ['*.cfg']
    }
)



import unittest2
import os
import tempfile
import shutil
import pip
from hoplite.client.status_updater import MockStatusUpdater
from hoplite.builtin_plugins.constants import InstallPythonPackageJobConstants as KEYS
from hoplite.builtin_plugins import install_python_package_job
from httmock import urlmatch, response, HTTMock

# Monkey patch pip so it doesn't mess with logging. Otherwise, presence of nose xunit logging handlers will cause an
# error when pip tries to set logging things
def blank_func(blank_arg):
    pass
import pip.basecommand
pip.basecommand.__dict__['logging_dictConfig'] = blank_func


@urlmatch(path='/reload$')
def reload_site_packages(url, request):
    return response(200)


class TestInstallPythonPackage(unittest2.TestCase):
    def test_install_from_local_path(self):
        setup_str = "from setuptools import setup, find_packages;setup(name='poopy', version='0.1', packages=find_packages())"
        tempdir = tempfile.mkdtemp()
        try:
            setup_py = open(os.path.join(tempdir, "setup.py"), 'w')
            setup_py.write(setup_str)
            setup_py.close()

            package_path = os.path.join(tempdir, "poopy")
            os.mkdir(package_path)

            init_file = open(os.path.join(package_path, "__init__.py"), 'w')
            init_file.close()

            config = {KEYS.LOCAL_PATH: tempdir}
            status = MockStatusUpdater()
            with HTTMock(reload_site_packages):
                install_python_package_job.run(config, status)
            self.assertTrue(status.status["succeeded"])
            try:
                import poopy
            except ImportError:
                self.fail("Could not import installed package")
        finally:
            pip.main(['uninstall', '-y', "poopy"])
            shutil.rmtree(tempdir)

    def test_install_fails_success_false_stdout_info(self):
        setup_str = "raise ValueError('I FAILED!')"
        tempdir = tempfile.mkdtemp()
        try:
            setup_py = open(os.path.join(tempdir, "setup.py"), 'w')
            setup_py.write(setup_str)
            setup_py.close()

            config = {KEYS.LOCAL_PATH: tempdir}
            status = MockStatusUpdater()


            install_python_package_job.run(config, status)

            self.assertFalse(status.status["succeeded"])

            # Because we monkey patch pip so it doesn't mess up nose xunit logging, the traceback info goes to the
            # console rather than to status.status.stdout
            #self.assertRegexpMatches(status.status["stdout"], "Traceback")

            self.assertIn("Pip returned a non-zero error code", status.status["errors"])
        finally:
            shutil.rmtree(tempdir)

    def test_missing_local_path_returns_errors(self):
        config = {}
        status = MockStatusUpdater()

        install_python_package_job.run(config, status)

        self.assertFalse(status.status["succeeded"])
        self.assertIn("No local path specified", status.status["errors"])


# Note - only certain exceptions can be pickled and unpickled. Specifically,
# custom exceptions with arguments to the __init__ function must be written
# with all arguments as optional, as shown below.


class ExternalCustomError(Exception):
    def __init__(self, msg=None):
        self.message = msg


class ExternalEmptyError(Exception):
    pass


import os
import shutil
import types

from hoplite.utils import server_logging
from hoplite.server.jobs.job_wrapper import job_wrapper
from hoplite.client.status_updater import MockStatusUpdater
from multiprocessing import Pipe
import tempfile
from tests import HopliteTestCase

logger = server_logging.get_server_logger(__name__)


class TestJobWrapper(HopliteTestCase):
    def test_job_wrapper_calls_run_on_passed_in_module(self):
        temp_dir = tempfile.mkdtemp()
        temp_file = os.path.join(temp_dir, "temp.txt")
        config = { "file_to_create": temp_file }
        module_name = self.test_jobs_module.constants.CREATE_FILE_JOB_NAME
        to_job, to_self = Pipe()
        job_wrapper(to_self, module_name, config, MockStatusUpdater(), entry_point_group_name="hoplite.test_jobs")
        try:
            self.assertTrue(os.path.isfile(temp_file))
        except Exception, e:
            raise e
        finally:
            shutil.rmtree(temp_dir)

    def test_job_wrapper_fills_pipe_with_exception_info(self):
        module_name = self.test_jobs_module.constants.THROW_AN_EXCEPTION_JOB_NAME
        config = {}
        to_job, to_self = Pipe()
        job_wrapper(to_self, module_name, config, MockStatusUpdater(), entry_point_group_name="hoplite.test_jobs")

        exec_info = to_job.recv()
        # Get to the bottom level of the exception information
        while ('type' not in exec_info) and (exec_info is not None):
            exec_info = exec_info.get('previous_exception', None)
        try:
            self.assertEqual(exec_info['type'], str(TypeError))
            self.assertEqual(exec_info['message'], "THE SKY IS FALLING!!")
        except Exception, e:
            raise e
        finally:
            to_job.close()
            to_self.close()

    def test_job_wrapper_fills_pipe_with_exception_info_bubble_up(self):
        module_name = self.test_jobs_module.constants.JOB_FAILED_EXCEPTION_JOB_NAME
        config = {}
        to_job, to_self = Pipe()
        job_wrapper(to_self, module_name, config, MockStatusUpdater(), entry_point_group_name="hoplite.test_jobs")

        exec_info = to_job.recv()
        exec_info = exec_info.get('previous_exception', None)
        try:
            self.assertEqual(exec_info['address'], "10.2.1.1")
            self.assertEqual(exec_info['uuid'], "5")
            self.assertIsInstance(exec_info['traceback'], types.TracebackType)
            # Get to the very bottom level of the exception information
            exec_info = exec_info.get('previous_exception', None)
            self.assertEqual(exec_info['message'], "Test Message")
            self.assertEqual(exec_info['type'], "Test Type String")
            self.assertEqual(exec_info['exception_object'], "pickled_string")
        except Exception, e:
            raise e
        finally:
            to_job.close()
            to_self.close()


import pickle
import sys
from tblib import pickling_support
import traceback

from hoplite.utils import server_logging
from hoplite.plugin_manager import EntryPointManager
from hoplite.exceptions import JobFailedError

# This makes it so that traceback objects can be pickled
pickling_support.install()


def job_wrapper(pipe_to_parent, entry_point_name, config, status_updater,
                entry_point_group_name='hoplite.jobs', uuid=''):
    """
    A picklable function that is used to start the job. It loads the specified
    module and calls run on it with the correct parameters.

    In the event of an error occurring in the job, the error is bubbled up to
    the highest parent of the job. This is done by encapsulating each exception
    into a JobFailedError, which is raised to cause the bubbling action. Since
    Exceptions aren't picklable, information from previous exceptions is put
    into a dictionary.

    This exception bubbling is useful for situations in which jobs are used to
    call other jobs. The stack trace for each "level" is saved, and the entire
    list of jobs with their respective traces can be displayed at the top level
    (where the JobFailedError is handled).
    """
    module = EntryPointManager(
        entry_point_group_name).get_plugin_module_by_name(entry_point_name)
    logger = server_logging.get_job_logger(module.__name__, uuid)
    try:
        module.run(config, status_updater)
    except JobFailedError as e:
        logger.error(
            "A job raised an exception and it was not caught."
            " Address: {0} UUID: {1}".format(e.addr, e.uuid))
        logger.error(
            "Exception Traceback: {0}".format(
                traceback.format_tb(e.traceback_object)))
        _, _, tb = sys.exc_info()
        traceback_object = tb
        exception_dictionary = {
            "address": e.addr,
            "uuid": e.uuid,
            "traceback": e.traceback_object,
            "previous_exception": e.previous_exception
        }
        pass_to_parent = {
            "traceback": traceback_object,
            "previous_exception": exception_dictionary
        }
        pipe_to_parent.send(pass_to_parent)
    except Exception as e:
        except_type, except_class, tb = sys.exc_info()
        traceback_object = tb
        type_string = str(except_type)
        try:
            pickled_exception = pickle.dumps(e)
        except pickle.PicklingError:
            pickled_exception = None

        exception_dictionary = {
            "type": type_string,
            "message": e.message,
            "exception_object": pickled_exception
        }
        pass_to_parent = {
            "traceback": pickle.dumps(traceback_object),
            "previous_exception": exception_dictionary
        }
        logger.error("Job UUID:{0} Type:{1} Finished with except type:{2} "
                     "except class:{3} traceback:{4}".format(
                        uuid, entry_point_name,
                        except_type, except_class, traceback.format_tb(tb)))
        pipe_to_parent.send(pass_to_parent)
    logger.debug("Finished running UUID:{0}".format(uuid))
